"""
AWS Secrets Manager
"""
import datetime
import os
import uuid
from decimal import Decimal
from pathlib import Path
import marshmallow as ma

from .util import AttrDict
from .parser import SecretParser
from . import fields


class NotSet:
    """
    NotSet is a placeholder for when a value is not set
    """
    pass


class SecretManager(SecretParser):
    """
    Secret Manager resolves secrets from AWS Secrets Manager
    This inherits from SecretParser that handles connecting to AWS and parsing the secret(s)

    Methods for type casting values:
    * `str`: String - This is the default type
    * `int`: Integer
    * `float`: Float
    * `decimal`: decimal.Decimal
    * `bool`: Boolean
    * `list`: List - You can specify a `delimiter` and a `subcast` type for list elements
    * `choices`: List[tuple] - You can specify a `delimiter` and a `subcast` type for list elements.
        Returns the form: [(key, value), (key, value)]
    * `datetime`: datetime.datetime - You can specify a `format` for the datetime string
    * `date`: datetime.date - You can specify a `format` for the date string
    * `time`: datetime.time - You can specify a `format` for the time string
    * `timedelta`: datetime.timedelta - You can specify a `format` for the timedelta string
    * `timedelta_seconds`: datetime.timedelta - Parses from seconds
    * `uuid`: uuid.UUID - You can specify a `version` for the UUID. 1=Time-based, 3=Name-based, 4=Random, 5=Name-based
    * `log_level`: int - Parses a log level string to an int
    * `path`: pathlib.Path - Parses a string to a pathlib.Path object
    * `dict`: AttrDict - You can specify a `prefix` for the dictionary keys, a `subcast_keys` type,
        and a `subcast_values` type
    """

    def value(self, name, default=NotSet) -> str:
        """
        Get the value of a secret.  Returns raw format.
        """
        if not self._secrets:
            self.load()
        # Try to get the value from secrets
        for secret in reversed(self._secrets.values()):  # We want to get the most recent secret first
            try:
                return secret.SecretValues[name]
            except KeyError:
                pass
        # Try to get value from environment variables
        if name in os.environ:
            return os.environ[name]
        elif name.upper() in os.environ:
            return os.environ[name.upper()]

        if default is not NotSet:
            return default

        # No value found
        raise KeyError(f'Key "{name}" is not found.')

    def str(self, name, default: (str, NotSet) = NotSet) -> str:
        """
        Get the value of a secret as a string
        """
        return fields.Str().deserialize(self.value(name, default=default))

    def int(self, name, default: (str, NotSet) = NotSet) -> int:
        """
        Get the value of a secret as an integer
        """
        return fields.Int().deserialize(self.value(name, default=default))

    def float(self, name, default: (str, NotSet) = NotSet) -> float:
        """
        Get the value of a secret as a float
        """
        return fields.Float().deserialize(self.value(name, default=default))

    def decimal(self, name, default: (str, NotSet) = NotSet) -> Decimal:
        """
        Get the value of a secret as a decimal
        """
        return fields.Decimal().deserialize(self.value(name, default=default))

    def bool(self, name, default: (str, NotSet) = NotSet) -> bool:
        """
        Get the value of a secret as a boolean
        """
        return fields.Bool().deserialize(self.value(name, default=default))

    def list(self, name, delimiter=',', subcast=fields.Str, default: (str, NotSet) = NotSet) -> list:
        """
        Get the value of a secret as a list
        """
        if isinstance(subcast, type):
            subcast = subcast()
        return [subcast.deserialize(v) for v in self.value(name, default=default).split(delimiter)]

    def choices(self, name, delimiter=',', subcast: ma.fields.Field = fields.Str,
                default: (str, NotSet) = NotSet) -> list:
        """
        Get the value of a secret as a list of tuples
        """
        return fields.Choices(delimiter=delimiter, subcast=subcast).deserialize(self.value(name, default=default))

    def datetime(self, name, format='%Y-%m-%d %H:%M:%S', default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a datetime
        """
        return fields.Datetime(format=format).deserialize(self.value(name, default=default))

    def date(self, name, format='%Y-%m-%d', default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a date
        """
        return fields.Date(format=format).deserialize(self.value(name, default=default))

    def time(self, name, format='%H:%M:%S', default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a time
        """
        return fields.Time(format=format).deserialize(self.value(name, default=default))

    def timedelta(self, name, default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a timedelta
        Format: HH:MM:SS
        """
        value = self.value(name, default=default)
        return datetime.timedelta(**{k: int(v) for k, v in zip(['hours', 'minutes', 'seconds'], value.split(':'))})

    def timedelta_seconds(self, name, default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a timedelta
        """
        return datetime.timedelta(seconds=self.int(name, default=default))

    def uuid(self, name, version=4, default: (str, NotSet) = NotSet) -> uuid:
        """
        Get the value of a secret as a UUID
        """
        return fields.UUID(metadata=dict(version=version)).deserialize(self.value(name, default=default))

    def log_level(self, name, default: (str, NotSet) = NotSet) -> int:
        """
        Get the value of a secret as a log level
        """
        return fields.LogLevel().deserialize(self.value(name, default=default))

    def path(self, name, default: (str, NotSet) = NotSet) -> Path:
        """
        Get the value of a secret as a Path
        """
        return fields.Path().deserialize(self.value(name, default=default))

    def _parse_dict(self, prefix, dictionary: dict, response_dict: AttrDict = None,
                    subcast_keys: ma.fields.Field = fields.Str,
                    subcast_values: ma.fields.Field = fields.Str) -> AttrDict:
        """
        Parse a dictionary from a dictionary of any type.

        WILL NOT OVERRIDE EXISTING KEYS. If a key already exists, it will be skipped.

        :param prefix: Filter prefix
        :param dictionary: Dictionary to parse
        :param subcast_keys: The type to cast the keys to
        :param subcast_values:  The type to cast the values to
        :return: AttrDict
        """
        for _key, _value in dictionary.items():
            if _key.startswith(prefix) or _key.startswith(prefix.upper()):
                # Remove the prefix from the key
                clean_key = _key.split('__', 1)[1]
                subkeys = clean_key.split('__')
                _dict = response_dict
                for subkey in subkeys[:-1]:
                    subkey = subcast_keys.deserialize(subkey)
                    if subkey not in _dict:
                        _dict[subkey] = AttrDict()
                    _dict = _dict[subkey]

                final_key = subcast_keys.deserialize(subkeys[-1])
                if final_key in _dict:
                    continue
                _dict[final_key] = subcast_values.deserialize(_value)
        return response_dict

    def dict(self, prefix, subcast_keys: ma.fields.Field = fields.Str,
             subcast_values: ma.fields.Field = fields.Str) -> AttrDict:
        """
        Get the value of a secret as a dictionary.

        # Key format
        DATABASE__default__HOST=localhost
        DATABASE__default__PORT=5432
        DATABASE__default__USER=postgres
        DATABASE__default__PASSWORD=postgres
        DATABASE__default__NAME=postgres
        DATABASE__default__OPTIONS__sslmode=disable

        # Returns
        {
            'default': {
                'HOST': 'localhost',
                'PORT': 5432,
                'USER': 'postgres',
                'PASSWORD': 'postgres',
                'NAME': 'postgres',
                'OPTIONS': {
                    'sslmode': 'disable'
                }
            }
        }
        """
        if isinstance(subcast_keys, type):
            subcast_keys = subcast_keys()
        if isinstance(subcast_values, type):
            subcast_values = subcast_values()

        if not self._secrets:
            self.load()
        filter_prefix = f'{prefix}__'

        response = AttrDict()
        # Load all values that begin with prefix
        for secret in reversed(self._secrets.values()):  # We want to get the most recent secret first
            response.update(self._parse_dict(filter_prefix, secret.SecretValues, response,
                                             subcast_keys, subcast_values))

        # Load all environment variables that begin with prefix
        response.update(self._parse_dict(filter_prefix, os.environ, response, subcast_keys, subcast_values))

        return response



"""
Test fields.py
"""
import datetime
import decimal
import logging
import pathlib
import unittest

import marshmallow.exceptions

from supersecret import fields


class TestFields(unittest.TestCase):
    def test_str(self):
        """
        Test the Str field class.
        """
        field = fields.Str()
        self.assertEqual(field.deserialize('test'), 'test')
        self.assertEqual(field.deserialize(u'test'), 'test')

    def test_int(self):
        """
        Test the Int field class.
        """
        field = fields.Int()
        self.assertEqual(field.deserialize('1'), 1)
        self.assertEqual(field.deserialize(u'1'), 1)

    def test_float(self):
        """
        Test the Float field class.
        """
        field = fields.Float()
        self.assertEqual(field.deserialize('1.1'), 1.1)
        self.assertEqual(field.deserialize(u'1.1'), 1.1)

    def test_decimal(self):
        """
        Test the Decimal field class.
        """
        field = fields.Decimal()
        self.assertEqual(field.deserialize('1.1'), decimal.Decimal('1.1'))
        self.assertEqual(field.deserialize(u'1.1'), decimal.Decimal('1.1'))

    def test_bool(self):
        """
        Test the Bool field class.
        """
        field = fields.Bool()
        self.assertEqual(field.deserialize('True'), True)
        self.assertEqual(field.deserialize(u'True'), True)
        self.assertEqual(field.deserialize('true'), True)
        self.assertEqual(field.deserialize(u'true'), True)
        self.assertEqual(field.deserialize('1'), True)
        self.assertEqual(field.deserialize('False'), False)
        self.assertEqual(field.deserialize(u'False'), False)
        self.assertEqual(field.deserialize('false'), False)
        self.assertEqual(field.deserialize(u'false'), False)
        self.assertEqual(field.deserialize('0'), False)

    def test_choices(self):
        """
        Test the Choices field class.
        """
        field = fields.Choices()
        self.assertEqual(field.deserialize('a:1,b:2'), [('a', '1'), ('b', '2')])
        self.assertEqual(field.deserialize(u'a:1,b:2'), [('a', '1'), ('b', '2')])
        self.assertEqual(field.deserialize([('a', '1'), ('b', '2')]), [('a', '1'), ('b', '2')])

    def test_timedelta(self):
        """
        Test TimeDelta field class.
        """
        field = fields.TimeDelta()
        self.assertEqual(field.deserialize('1:0:0'), datetime.timedelta(hours=1))


    def test_timedeltaseconds(self):
        """
        Test the TimeDeltaSeconds field class.
        """
        field = fields.TimeDeltaSeconds()
        self.assertEqual(field.deserialize('1'), datetime.timedelta(seconds=1))

    def test_path(self):
        """
        Test the Path field class.
        """
        field = fields.Path()
        self.assertEqual(field.deserialize('/tmp'), pathlib.Path('/tmp'))
        self.assertEqual(field.deserialize(u'/tmp'), pathlib.Path('/tmp'))
        self.assertEqual(field.deserialize(pathlib.Path('/tmp')), pathlib.Path('/tmp'))


    def test_loglevel(self):
        """
        Test the LogLevel field class.
        """
        field = fields.LogLevel()
        self.assertEqual(field.deserialize('DEBUG'), logging.DEBUG)
        self.assertEqual(field.deserialize('INFO'), logging.INFO)
        self.assertEqual(field.deserialize('WARNING'), logging.WARNING)
        self.assertEqual(field.deserialize('ERROR'), logging.ERROR)
        self.assertEqual(field.deserialize('CRITICAL'), logging.CRITICAL)

        # Test invalid log levels
        with self.assertRaises(marshmallow.exceptions.ValidationError):
            field.deserialize('INVALID')



"""
Test other parts of the supersecret package.
"""

import unittest

from supersecret import get_version
from supersecret.util import AttrDict


class TestOther(unittest.TestCase):
    def test_version(self):
        self.assertEqual(len(get_version().split('.')), 3)

    def test_attrdict(self):
        """
        Test the AttrDict class.
        """
        d = AttrDict()
        d['test'] = 'test'
        self.assertEqual(d.test, 'test')

        d.test = 'test2'
        self.assertEqual(d['test'], 'test2')

        d = AttrDict({'test': 'test'})
        self.assertEqual(d.test, 'test')

        del d.test
        self.assertEqual(dict(d), {})



import dataclasses
from dataclasses import dataclass
from typing import List, Dict
from datetime import datetime

from supersecret.util import AttrDict


def nested_dataclass(*args, **kwargs):
    def wrapper(cls):
        cls = dataclass(cls, **kwargs)
        original_init = cls.__init__

        def __init__(self, *args, **kwargs):
            for name, value in kwargs.items():
                field_type = cls.__annotations__.get(name, None)
                try:
                    if field_type.__dataclass_fields__.get('data'):
                        new_obj = field_type(data=value)
                        kwargs[name] = new_obj

                    else:
                        raise AttributeError

                except AttributeError:
                    if dataclasses.is_dataclass(field_type) and isinstance(value, dict):
                        new_obj = field_type(**value)
                        kwargs[name] = new_obj

            original_init(self, *args, **kwargs)

        cls.__init__ = __init__
        return cls

    return wrapper(args[0]) if args else wrapper


@dataclass
class SecretValues:
    """
    The class signature is unknown and depends on the returned data
    """
    data: dataclasses.field(default_factory=AttrDict) = None

    def __post_init__(self):
        [setattr(self, k, v) for k, v in self.data.items()]

    def __getattr__(self, item):
        if item in self.data:
            return self.data[item]
        raise KeyError(f'Key "{item}" is not found.')

    def __getitem__(self, item):
        return self.data[item]

    def items(self):
        return self.data.items()


@dataclass
class MetadataDTO:
    pass


@dataclass
class ResponseMetadata:
    HTTPHeaders: Dict
    HTTPStatusCode: int
    RequestId: str
    RetryAttempts: int


@nested_dataclass
@dataclass
class GetValue:
    ARN: str
    Name: str
    VersionId: str
    SecretValues: SecretValues
    VersionStages: List[str]
    CreatedDate: datetime
    ResponseMetadata: ResponseMetadata



"""
Tests for the supersecret.manager module.
"""
import datetime
import decimal
import unittest
import json
import uuid
from collections import OrderedDict
from unittest.mock import patch
import os

from botocore.client import BaseClient

from supersecret.manager import SecretManager
from supersecret.util import AttrDict

SECRETS_MOCK = OrderedDict()
SECRETS_MOCK['TestingSecret'] = {
    'ARN': 'arn:aws:secretsmanager:us-east-1:123456789:secret:TestingSecret-123456',
    'Name': 'TestingSecret',
    'VersionId': '123456',
    'SecretString': json.dumps({
        'username': 'test_username',
        'password': 'test_password',
        'database__host': 'localhost',
        'database__port': '1234',
        'database__name': 'test_database',
        'database__username': 'test_database_username',
        'database__password': 'test_database_password',
        'database__options__ssl': 'True',
        'test_float': '1.234',
        'test_int': '1234',
        'test_bool': 'True',
        'test_list': 'test1,test2,test3',
        'test_choices': 'test1:test2,test3:test4',
        'test_datetime': '2020-01-01 00:00:00',
        'test_date': '2020-01-01',
        'test_decimal': '1.234',
        'test_time': '00:00:00',
        'test_timedelta': '1:00:00',
        'test_timedelta_seconds': '3600',
        'test_uuid': '12345678-1234-5678-1234-567812345678',
    }),
    'VersionStages': ['AWSCURRENT'],
    'CreatedDate': datetime.datetime.now(),
    'ResponseMetadata': {'RequestId': 'abc123',
                         'HTTPStatusCode': 200,
                         'HTTPHeaders': {'x-amzn-requestid': 'abc123',
                                         'content-type': 'application/x-amz-json-1.1',
                                         'content-length': '2033',
                                         'date': 'Sun, 26 Feb 2023 07:44:38 GMT'},
                         'RetryAttempts': 0}}

SECRETS_MOCK['TestingSecret2'] = {
    'ARN': 'arn:aws:secretsmanager:us-east-1:123456789:secret:TestingSecret-123457',
    'Name': 'TestingSecret2',
    'VersionId': '123457',
    'SecretString': json.dumps({
        'username': 'new_username',
        'password': 'new_password',
        'database__host': 'remote_host',
        'database__port': '5678',
        'database__name': 'new_database',
        'database__username': 'new_username',
        'database__password': 'new_database_password',
        'database__options__ssl': 'False',
        'test_float': '5.678',
        'test_int': '5678',
        'test_bool': 'False',
        'test_list': 'test4,test5,test6',
        'test_choices': 'test5:test6,test7:test8',
        'test_datetime': '2020-01-02 00:00:00',
        'test_date': '2020-01-02',
        'test_decimal': '5.678',
        'test_time': '01:00:00',
        'test_timedelta': '1:01:00',
        'test_timedelta_seconds': '3660',
        'test_uuid': '12345678-1234-5678-1234-567812345679',
    }),
    'VersionStages': ['AWSCURRENT'],
    'CreatedDate': datetime.datetime.now(),
    'ResponseMetadata': {'RequestId': 'abc123',
                         'HTTPStatusCode': 200,
                         'HTTPHeaders': {'x-amzn-requestid': 'abc123',
                                         'content-type': 'application/x-amz-json-1.1',
                                         'content-length': '2033',
                                         'date': 'Sun, 26 Feb 2023 07:44:38 GMT'},
                         'RetryAttempts': 0}}


class MockSecretsClient:
    """
    Mock class for the boto3 client.
    """

    def get_secret_value(self, SecretId):
        """
        Mock method for the get_secret_value method.
        """
        return AttrDict(SECRETS_MOCK[SecretId])


class TestSecretManager(unittest.TestCase):
    """
    Tests for the SecretManager class.
    """

    def test_get_value_env(self):
        """
        Test that the get_value method returns the expected value.
        """
        os.environ['test'] = 'test'
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.str('test'), 'test')

            secret = next(iter(secret_manager._secrets.values()))
            self.assertEqual(secret.SecretValues.username, 'test_username')
            self.assertEqual(secret.SecretValues['username'], 'test_username')

            self.assertRaises(KeyError, getattr, secret.SecretValues, 'does_not_exist')

    def test_connect_to_session(self):
        """
        Test that the connect_to_session method returns the expected value.
        """
        secret_manager = SecretManager('TestingSecret')
        self.assertIsInstance(secret_manager.connect(), BaseClient)

    def test_get_str(self):
        """
        Test that the get_str method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.str('username'), 'test_username')

    def test_get_int(self):
        """
        Test that the get_int method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.int('test_int'), 1234)

    def test_get_float(self):
        """
        Test that the get_float method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.float('test_float'), 1.234)

    def test_get_bool(self):
        """
        Test that the get_bool method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertTrue(secret_manager.bool('test_bool'))

    def test_get_list(self):
        """
        Test that the get_list method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.list('test_list'), ['test1', 'test2', 'test3'])

    def test_get_choices(self):
        """
        Test that the get_choices method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.choices('test_choices'), [('test1', 'test2'), ('test3', 'test4')])

    def test_get_datetime(self):
        """
        Test that the get_datetime method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.datetime('test_datetime'), datetime.datetime(2020, 1, 1, 0, 0))

    def test_get_date(self):
        """
        Test that the get_date method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.date('test_date'), datetime.date(2020, 1, 1))

    def test_get_decimal(self):
        """
        Test that the get_decimal method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.decimal('test_decimal'), decimal.Decimal('1.234'))

    def test_get_time(self):
        """
        Test that the get_time method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.time('test_time'), datetime.time(0, 0))

    def test_get_timedelta(self):
        """
        Test that the get_timedelta method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.timedelta('test_timedelta'), datetime.timedelta(seconds=3600))

    def test_get_uuid(self):
        """
        Test that the get_uuid method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.uuid('test_uuid'), uuid.UUID('12345678-1234-5678-1234-567812345678'))

    def test_default(self):
        """
        Test that the default method returns the expected value.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            self.assertEqual(secret_manager.str('test_default', default='test'), 'test')
            self.assertEqual(secret_manager.int('test_default', default=123), 123)

    def test_dict(self):
        """
        Test that the dict method returns the expected value.
        """
        desired_response = AttrDict({
            'host': 'localhost',
            'port': '1234',
            'name': 'test_database',
            'username': 'test_database_username',
            'password': 'test_database_password',
            'options': AttrDict({'ssl': 'True'}),
        })
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager('TestingSecret')
            secret_dict = secret_manager.dict('database')
            self.assertEqual(secret_dict, desired_response)

    def test_multiple_keys(self):
        """
        Create a manager with multiple keys and test that the correct value is returned.
        """
        first_secret = list(SECRETS_MOCK.keys())[0]
        last_secret = list(SECRETS_MOCK.keys())[-1]
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            secret_manager = SecretManager(first_secret)
            secret_manager.load(last_secret)
            self.assertEqual(secret_manager.str('test'), 'test')
            self.assertNotEqual(secret_manager.str('username'), 'test_username')
            self.assertEqual(secret_manager.str('username'), 'new_username')

    def test_environment_key_name(self):
        """
        Test using SECRET_NAME as the environment variable name to set the key name.
        """
        os.environ['SECRET_NAME'] = 'TestingSecret2'
        try:
            with patch.object(SecretManager, 'connect') as mock_connect_to_session:
                mock_connect_to_session.return_value = MockSecretsClient()
                secret_manager = SecretManager()
                self.assertEqual(secret_manager.str('password'), 'new_password')
        finally:
            del os.environ['SECRET_NAME']

    def test_context_manager_usage(self):
        """
        Test that the context manager usage works as expected.
        We mock the connect method to return a mock secrets client.
        We are unable to test the client status or close because the method is mocked.
        We do this test in the actual AWS connection tests.
        """
        with patch.object(SecretManager, 'connect') as mock_connect_to_session:
            mock_connect_to_session.return_value = MockSecretsClient()
            with SecretManager('TestingSecret') as secret_manager:
                self.assertEqual(secret_manager.str('username'), 'test_username')
                self.assertEqual(secret_manager.str('password'), 'test_password')



"""
Marshmallow fields for supersecret.

Fields:
* `Str`: String - This is the default type
* `Int`: Integer
* `Float`: Float
* `Decimal`: decimal.Decimal
* `Bool`: Boolean
* `List`: List - You can specify a `delimiter` and a `subcast` type for list elements
* `Choices`: List[tuple] - You can specify a `delimiter` and a `subcast` type for list elements.
        Returns the form: [(key, value), (key, value)]
* `Datetime`: datetime.datetime - You can specify a `format` for the datetime string
* `Date`: datetime.date - You can specify a `format` for the date string
* `Time`: datetime.time - You can specify a `format` for the time string
* `TimeDelta`: datetime.timedelta - You can specify a `format` for the timedelta string
* `UUID`: uuid.UUID - You can specify a `version` for the UUID. 1=Time-based, 3=Name-based, 4=Random, 5=Name-based
* `LogLevel`: int - Parses a log level string to an int
* `Path`: pathlib.Path - Parses a string to a pathlib.Path object

"""
import datetime
import logging
import pathlib

import marshmallow as ma

Str = ma.fields.Str
Int = ma.fields.Int
Float = ma.fields.Float
Decimal = ma.fields.Decimal
Bool = ma.fields.Bool
List = ma.fields.List


class Choices(ma.fields.List):
    """
    A field that parses a string to a list of choices.
    Choices should be stored as: key:value,key:value
    Choices are returned as: [(key, value), (key, value)]
    """
    def __init__(self, delimiter=',', subcast=Str, *args, **kwargs):
        super().__init__(subcast, *args, **kwargs)
        self.delimiter = delimiter

    def _deserialize(self, value, *args, **kwargs) -> list:
        if isinstance(value, list):
            return value
        subcast = self.inner
        return [(v[0], subcast.deserialize(v[1])) for v in [v.split(':') for v in value.split(self.delimiter)]]


Datetime = ma.fields.DateTime
Date = ma.fields.Date
Time = ma.fields.Time


class TimeDelta(ma.fields.Str):
    """
    A field that parses a string to a datetime.timedelta object.
    """
    def _deserialize(self, value, *args, **kwargs) -> datetime.timedelta:
        if isinstance(value, datetime.timedelta):
            return value
        ret = super()._deserialize(value, *args, **kwargs)
        hours, minutes, seconds = ret.split(':')
        return datetime.timedelta(hours=int(hours), minutes=int(minutes), seconds=int(seconds))


TimeDeltaSeconds = ma.fields.TimeDelta
UUID = ma.fields.UUID


class Path(ma.fields.Str):
    """
    A field that parses a string to a pathlib.Path object.
    """
    def _deserialize(self, value, *args, **kwargs) -> pathlib.Path:
        if isinstance(value, pathlib.Path):
            return value
        ret = super()._deserialize(value, *args, **kwargs)
        return pathlib.Path(ret)


class LogLevel(ma.fields.Int):
    """
    A field that parses a log level string to an int.
    """
    def _format_num(self, value) -> int:
        try:
            return super()._format_num(value)
        except (TypeError, ValueError) as error:
            value = value.upper()
            if hasattr(logging, value) and isinstance(getattr(logging, value), int):
                return getattr(logging, value)
            else:
                raise ma.ValidationError("Not a valid log level.") from error



"""
AWS Secrets Parser
"""
import base64
import json
import os
from collections import OrderedDict

from botocore.exceptions import ClientError

from .dto import GetValue
from .exceptions import define_error
from botocore.client import BaseClient


class SecretParser:
    """
    SecretParser connects to AWS and parses the secret
    """
    SERVICE_NAME: str = 'secretsmanager'

    def __init__(self, default_secret_name: str = None, **aws_kwargs):
        """
        :param secret_name: The AWS Secrets Manager secret name
        :param aws_kwargs: AWS connection kwargs for boto3.client
        """
        if default_secret_name is None:
            default_secret_name = os.environ.get('SECRET_NAME', None)
        if default_secret_name is None:
            raise KeyError('No secret name provided')

        self.client = None
        self.__client_created = False
        self.default_secret_name = default_secret_name

        self.aws_kwargs = aws_kwargs
        self._secrets = OrderedDict()

    def connect(self) -> BaseClient:
        """
        Connect to AWS Secrets Manager to the default aws session
        :return: client
        """
        if self.client:
            return self.client
        import boto3
        self.aws_kwargs.setdefault('region_name', os.environ.get('AWS_REGION', 'us-east-1'))

        self.client = boto3.client(self.SERVICE_NAME, **self.aws_kwargs)
        self.__client_created = True
        return self.client

    def load(self, secret_name: str = None, client: BaseClient = None) -> GetValue:
        """
        Load secret from AWS Secrets Manager
        """
        if secret_name is None:
            secret_name = self.default_secret_name
        if secret_name in self._secrets:
            return self._secrets[secret_name]
        if client is None:
            client = self.client
        if client is None:
            client = self.connect()
        response = self._load_secret(secret_name, client)
        self._secrets[secret_name] = response
        return response

    def _load_secret(self, secret_name: str, client: BaseClient) -> GetValue:
        try:
            raw_secret = client.get_secret_value(
                SecretId=secret_name
            )
        except ClientError as error:
            error_code = error.response['Error']['Code']
            err = define_error(error_code)
            str(error)
            raise error

        else:
            if raw_secret.get('SecretBinary'):
                raw_secret['SecretValues'] = base64.b64decode(raw_secret['SecretBinary'])
                del raw_secret['SecretBinary']
            else:
                value = raw_secret['SecretString']
                if isinstance(value, str):
                    value = json.loads(value)
                raw_secret['SecretValues'] = value
                del raw_secret['SecretString']

            return GetValue(**raw_secret)

    def close(self):
        try:
            if self.__client_created and self.client:
                self.client.close()
        finally:
            self.client = None
            self._secrets = OrderedDict()

    # Context Management
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()



import inspect
import sys


def define_error(response_error):
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            if obj.__name__ == response_error:
                return obj


class BaseSecretsManagerException(Exception):
    """
    Base exception for all Secrets Manager exceptions
    """

    def __str__(self):
        return self.__doc__


class DecryptionFailureException(BaseSecretsManagerException):
    """
    Secrets Manager can't decrypt the protected secret text using the provided KMS key.
    """


class InternalServiceErrorException(BaseSecretsManagerException):
    """
    An error occurred on the server side.
    """


class InvalidParameterException(BaseSecretsManagerException):
    """
    You provided an invalid value for a parameter.Deal with the exception here, and/or rethrow at your discretion.
    """


class InvalidRequestException(BaseSecretsManagerException):
    """
    You provided a parameter value that is not valid for the current state of the resource.
    """


class ResourceNotFoundException(BaseSecretsManagerException):
    """
    We can't find the resource that you asked for. Deal with the exception here, and/or rethrow at your discretion.
    """



"""
This is a testing set to actually connect to AWS and test
pulling and parsing secrets from AWS Secrets Manager.
This is not a unit test, but a functional test.

Only runs if environment variable AWS_TESTING is set to True.

Your AWS IAM user must have the following permissions:
```json
{
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "VisualEditor0",
            "Effect": "Allow",
            "Action": [
                "secretsmanager:*",
                "secretsmanager:DeleteSecret"
            ],
            "Resource": "*"
        }
    ]
}
```


"""

import unittest
import os
import json
import uuid
from collections import OrderedDict

import boto3
from botocore.exceptions import ClientError

from supersecret.exceptions import BaseSecretsManagerException
from supersecret.manager import SecretManager
from .test_manager import SECRETS_MOCK

# Set up the AWS testing environment
AWS_TESTING = bool(os.environ.get('AWS_TESTING', False))

UUID = str(uuid.uuid4())

if AWS_TESTING:
    # Setup AWS Connection
    CLIENT = boto3.client('secretsmanager', region_name=os.environ.get('AWS_REGION', 'us-east-1'))
else:
    CLIENT = None


class TestAwsConnections(unittest.TestCase):
    """
    Tests for the SecretManager class.
    """
    secrets = OrderedDict()
    SECRET_NAMES = {f'{name}-{UUID}': name for name in SECRETS_MOCK.keys()}

    def setUp(self) -> None:
        """
        Set up the testing environment.
        """
        if not AWS_TESTING:
            self.skipTest('AWS_TESTING not set to True')

        if not self.secrets:
            self.create_secrets()

        for i, secret in enumerate(self.secrets.keys()):
            if i == 0:
                self.secret_manager = SecretManager(secret)
            else:
                self.secret_manager.load(secret)

    # Remove secret after ALL tests have run
    @classmethod
    def tearDownClass(cls) -> None:
        """
        Tear down the testing environment.
        """
        if cls.secrets:
            cls.delete_secrets()

    def create_secrets(self):
        """
        Create a test secret in AWS Secrets Manager.
        """
        for secret_name, secret in self.SECRET_NAMES.items():
            secret_string = SECRETS_MOCK['TestingSecret']['SecretString']
            CLIENT.create_secret(
                Name=secret_name,
                SecretString=secret_string
            )
            self.secrets[secret_name] = json.loads(secret_string)

    @classmethod
    def delete_secrets(cls):
        """
        Delete the test secret from AWS Secrets Manager.
        """
        for secret_name in cls.SECRET_NAMES.keys():
            CLIENT.delete_secret(SecretId=secret_name, ForceDeleteWithoutRecovery=True)

    def test_exceptions(self):
        """
        Test that the exceptions are raised as expected.
        """
        os.environ['test'] = 'test'

        self.assertRaises(ClientError, self.secret_manager.load, 'does_not_exist')

    def test_get_value_env(self):
        """
        Test that the get_value method returns the expected value.
        """
        os.environ['test'] = 'test'
        self.assertEqual(self.secret_manager.str('test'), 'test')

    def test_get_str(self):
        """
        Test that the get_str method returns the expected value.
        """
        # Should be the value from the last secret
        secret = self.secrets[list(self.secrets.keys())[-1]]
        self.assertEqual(self.secret_manager.str('username'), secret['username'])

    def test_get_int(self):
        """
        Test that the get_int method returns the expected value.
        """
        # Should be the value from the last secret
        secret = self.secrets[list(self.secrets.keys())[-1]]
        self.assertEqual(self.secret_manager.int('test_int'), int(secret['test_int']))

    def test_context_manager_usage(self):
        """
        Test that the context manager usage works as expected.
        We mock the connect method to return a mock secrets client.
        We are unable to test the client status or close because the method is mocked.
        We do this test in the actual AWS connection tests.
        """
        name = list(self.secrets.keys())[0]
        with SecretManager(name) as secret_manager:
            self.assertIsNone(secret_manager.client)
            self.assertEqual(secret_manager.str('username'), 'test_username')
            self.assertEqual(secret_manager.str('password'), 'test_password')
            self.assertIsNotNone(secret_manager.client)




class AttrDict(dict):
    """
    A dictionary that allows you to access keys as attributes.
    """
    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value

    def __delattr__(self, attr):
        del self[attr]

    def __repr__(self):
        return f'<AttrDict {super().__repr__()}>'
